****************************************************************************
*  User modifiable functions used in NT2/CA3 but dependent on hardware     *
*  or not allowed for a normal User on a Unix System.                      *
*  These functions cannot be implemented because they require capabilities *
*  of an IBM-PC or violate Unix protection mechanisms.                     *
*                                                                          *
*  Note: When FS_SET("devel", .T.) is active, you will receive an          *
*        additional iformation on unsupported functions.                   *
*                                                                          *
*  If you need to change this file, copy it into your local directory,     *
*  make changes and recompile it by using                                  *
*     FlagShip fs2_udfs.prg -c -m -na                                      *
*  and then link it with your application                                  *
*     FlagShip myprog*.prg fs2_udfs.o -lFS2                                *
*                                                                          *
*                           (c) 1992-96 multisoft Datentechnik GmbH Munich *
****************************************************************************

#Cinline
#include "FSopenc.h"
#endCinline

#include "set.fh"

static MSG_VIDEO  := "Video manipulation not possible, returning "
static MSG_FONT   := "Font manipulation not possible, returning "
static MSG_MEMORY := "Memory manipulation not possible, returning "
static MSG_PORT   := "Port manipulation not possible, returning "
static MSG_KEYB   := "Keyboard manipulation not possible, returning "
static MSG_HARDW  := "Hardware manipulation not possible, returning "
static MSG_DATE   := "Date/Time can only be set by su/root, returning "
static MSG_PRINT  := "Printer is usually managed by lpr spooler.;" + ;
                     "To set a com/lpt port, use com_*() functions. Returning "
static MSG_NOVEL  := "There is usually no connection to Novell;" + ;
                     "except via NFS (or IPX on some Unix systems),;" + ;
                     "nor are the Novell APIs available here. Returning "

***********************************************************
* Functions to access video-functions and / or videomodes *
***********************************************************


* Ŀ
*  cga40                              Mode cannot be changed, even on CGA  
* 
function cga40
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  cga80                              Mode cannot be changed, even on CGA  
* 
function cga80
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  ega43                              Mode cannot be changed, even on EGA  
* 
function ega43
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  vga28                              Mode cannot be changed, even on VGA  
* 
function vga28()
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  vga50                              Mode cannot be changed, even on VGA  
* 
function vga50
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  monochrome                         This videocard cannot be manipulated 
* 
function monochrome
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  monochrom                          This videocard cannot be manipulated 
* 
function monochrom
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

* Ŀ
*  getmode           Mode cannot be changed. Monochrome wont change it!    
* 
function getmode
fs2_devWarn(MSG_VIDEO + "monochrome")
return "monochrome"

* Ŀ
*  iscga                              This videocard cannot be manipulated 
* 
function iscga
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return iscolor() may make sense

* Ŀ
*  isega                              This videocard cannot be manipulated 
* 
function isega
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return iscolor() may make sense

* Ŀ
*  isvga                              This videocard cannot be manipulated 
* 
function isvga
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return iscolor() may make sense

* Ŀ
*  ishercules                         This videocard cannot be manipulated 
* 
function ishercules
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return !iscolor() may make sense

* Ŀ
*  ismono                             This videocard cannot be manipulated 
* 
function ismono
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return !iscolor() may make sense

* Ŀ
*  ispga                              This videocard cannot be manipulated 
* 
function ispga
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return iscolor() may make sense

* Ŀ
*  ismcga                             This videocard cannot be manipulated 
* 
function ismcga
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.         && return iscolor() may make sense

* Ŀ
*  setscrmode                         This videocard cannot be manipulated 
* 
function setscrmode
parameters p1
fs2_devWarn(MSG_VIDEO + ".F.")
return .F.

*************************************************************************
* functions to do fontmanipulations These functions require an EGA/VGA  *
* Adaptor. Fontmanipulation is possible with FlagShip directly, see     *
* FlagShip User Manual, chap 9-9 for a description of output character  *
* mapping.                                                              *
*************************************************************************


* Ŀ
*  fontload               We cannot load fonts independant of the terminal 
* 
function fontload
parameters p1, p2, p3, p4
fs2_devWarn(MSG_FONT + "-2")
return -2

* Ŀ
*  fontreset           Fontload did not change, so reset is not neccessary 
* 
function fontreset
parameters p1
fs2_devWarn(MSG_FONT + ".T.")
return .T.

* Ŀ
*  fontselect                                        No selection of fonts 
* 
function fontselect
parameters p1
fs2_devWarn(MSG_FONT + "-2")
return -2

* Ŀ
*  maxfont                                       Number of available fonts 
* 
function maxfont
fs2_devWarn(MSG_FONT + "1")
return 1


**************************************************
* Functions to access memory, I/O Ports and Bios *
**************************************************


* Ŀ
*  inbyte                         No permission to manipulate I/O Ports    
* 
function inbyte
parameters p1
fs2_devWarn(MSG_PORT + "-1;Use com_*() functions instead")
return -1

// Note: the handling of i/o ports in Unix is similar to handling usual
// files. The INBYTE() function can be read:
//
// local fd := Fopen("/dev/cua0", 0), cChar := " "
// if fd == 0
//   return -1
// endif
// Fread(fd, @cChar, 1)
// Fclose(fd)
// return asc(cChar)
//
// where the /dev/cua0 is a device for 1st serial port, /dev/lp0 for
// the parallel port etc. There are three handles always open: 1=stdin,
// 2=stdout and 3=stderr, so you don't need Fopen/Fclose them.
// As you may see, the above algorithm is VERY inefficient, becouse of 
// the open/read/close for every byte.  It is better to Fopen() the device
// once, Fread() as long as you want or until Ferror() etc, and then 
// close the device by Fclose().

* Ŀ
*  inword                         No permission to manipulate I/O Ports    
* 
function inword
parameters p1
fs2_devWarn(MSG_PORT + "-1;Use com_*() functions instead")
return -1

// Note: similar to INBYTE(), but two chars are read

* Ŀ
*  outbyte                        No permission to manipulate I/O Ports    
* 
function outbyte
parameters p1, p2
fs2_devWarn(MSG_PORT + ".F.;Use com_*() functions instead")
return .F.

// Note: the handling of i/o ports in Unix is similar to handling usual
// files. The OUTBYTE() function can be read:
//
// local fd := Fopen("/dev/cua0", 1)
// if fd == 0
//   return .F.
// endif
// Fwrite(fd, p2, 1)  // or Fwrite(fd, p2, len(p2)) for a string
// Fclose(fd)
// return .T.
//
// where the /dev/cua0 is a device for 1st serial port, /dev/lp0 for
// the parallel port etc. There are three handles always open: 1=stdin,
// 2=stdout and 3=stderr, so you don't need Fopen/Fclose them.
// As you may see, the above algorithm is VERY inefficient, becouse of 
// the open/read/close for every byte/string.  It is better to Fopen() the 
// device once, Fwrite() as long as you want or until Ferror() etc, and 
// then close the device by Fclose().

* Ŀ
*  outword                        No permission to manipulate I/O Ports    
* 
function outword
parameters p1, p2
fs2_devWarn(MSG_PORT + ".F.;Use com_*() functions instead")
return .F.

// Note: similar to OUTBYTE(), but two chars are written

* Ŀ
*  peekbyte                       No permission to manipulate Memory       
* 
function peekbyte
parameters p1, p2
fs2_devWarn(MSG_MEMORY + "-1")
return -1

* Ŀ
*  peekstr                        No permission to manipulate Memory       
* 
function peekstr
parameters p1, p2, p3
fs2_devWarn(MSG_MEMORY + "-1")
return -1

* Ŀ
*  peekword                       No permission to manipulate Memory       
* 
function peekword
parameters p1, p2
fs2_devWarn(MSG_MEMORY + "-1")
return -1

* Ŀ
*  pokebyte                       No permission to manipulate Memory       
* 
function pokebyte
parameters p1, p2, p3
fs2_devWarn(MSG_MEMORY + "-1")
return -1

* Ŀ
*  pokeword                       No permission to manipulate Memory       
* 
function pokeword
parameters p1, p2, p3
fs2_devWarn(MSG_MEMORY + "-1")
return -1

* Ŀ
*  dsetkbios                      Bios is unused or dont exist under UNIX  
* 
function dsetkbios
parameters p1
fs2_devWarn("No DOS-BIOS available on UNIX;Use system() calls instead.;Returning .F.")
return .F.

* Ŀ
*  kbdtype                 It is a terminal, its delivered codes depend on 
*                                the Terminfo description, not on Hardware 
* 
function kbdtype
fs2_devWarn(MSG_KEYB + "0")
return 0     && == PC-Keyboard

* Ŀ
*  kbdstat             It is a terminal, its status is unknown to the host 
* 
function kbdstat
fs2_devWarn(MSG_KEYB + "0")
return 0

* Ŀ
*  ksetcaps            It is a terminal, its status is unknown to the host 
*  ksetnum                                                                 
*            note that functions ksetins and ksetscroll are true functions:
*            ksetscroll() changes behaviour of cursor keys with windows    
*            ksetsins()   toggles the insert mode                          
* 
function ksetcaps
parameters p1
fs2_devWarn(MSG_KEYB + ".F.")
return .F.
function ksetnum
parameters p1
fs2_devWarn(MSG_KEYB + ".F.")
return .F.

* Ŀ
*  isat                  This is not even a PC normally, so we do not have 
*                         special AT-Type Hardware wich can be programmed  
* 
function isat
fs2_devWarn(MSG_HARDW + ".F.")
return .F.

* Ŀ
*  speed                This is not a real time Operating System, so there 
*                       is no fixed speed index                            
*                       Maybe, a speed index based on seconds() or         
*                       secondscpu() may make sense                        
* 
function speed
parameters p1
*
* xx = secondscpu()
* for ii = 1 to 100000
*   jj = ii + 1
* next
* xx -= secondscpu()
* return 45000 / xx   && the result value is at least compareable
*
fs2_devWarn(MSG_HARDW + "5000")
return 5000


* Ŀ
*  setdate /settime          The System clock can only be set by superuser 
* 
function setdate
parameters p1
fs2_devWarn(MSG_DATE + ".F.")
return .F.
function settime
parameters p1
fs2_devWarn(MSG_DATE + ".F.")
return .F.


********************************************
* Functions to access printer directly     *
********************************************


* Ŀ
*  printinit              No permission to manipulate directly the Printer 
* 
function printinit
parameters p1, p2
fs2_devWarn(MSG_PRINT + "0")
return 0            && return no error

* Ŀ
*  printready       We normally write to file, the file is always ready    
* 
function printready
parameters p1
fs2_devWarn(MSG_PRINT + ".T.")
return .T.          && return printer is ready

* Ŀ
*  printfile        not available in CA3. Prints specified file            
* 
function printfile (name, clearHigh)
   local buff, handle
   local currCons  := set(_SET_CONSOLE)
   local currPrint := set(_SET_PRINTER)
   clearHigh := if (valtype(clearHigh) == "L", clearHigh, .F.)
   ** if !clearHigh
   **    RUN ("lp " + name)    // upon request
   ** else ...
   handle = fopen(name)
   if handle <= 1
      return .F.
   endif
   set console off
   set printer on
   while ferror() == 0
      buff := freadtxt(handle)
      if clearHigh
#Cinline
      {
         FSuchar *str;
         str = VAR_CHR (VAR_NAME_LOCAL(buff));
         for (; *str; str++)
            (*str) &= 0x7F;
      }
#endCinline
      endif
      ? buff
   enddo
   fclose (handle)
   set(_SET_CONSOLE, currCons)
   set(_SET_PRINTER, currPrint)
   return .T.

* Ŀ
*  printerror       can be simulated only                                  
* 
function printerror()
   return doserror()

* Ŀ
*  printscr([p1],[p2],[p3])   print screen content                         
*                                                                          
*  [p1]  = .T. print row number, .F. content only (default)                
*  [p2]  = file name (or driver), if not given = std.printer output file   
*  [p3]  = .T. line ends with CR/LF, .F. with LF only (default)            
*                                                                          
*  Note: the characters mapped by FSchrmap.def on the screen may cause     
*        other printer output, since no reverse mapping                    
* 
function printscr(printRow, printFile, DOSline)
   LOCAL_INT ymax := MAXROW()                       // used also in C
   LOCAL     chr_arr[ymax +1]                       // e.g. chr_arr[25]
   LOCAL     ii := 0, CrLf, dummy
   LOCAL     lSetPrint := SET(_SET_PRINTER)
   LOCAL     cSetPrint := SET(_SET_PRINTFILE)
   LOCAL     lSetCons  := SET(_SET_CONSOLE)
   LOCAL     nRow := ROW(), nCol := COL()

   printRow  := if (valtype(printRow ) == "L", printRow , .F.)
   printFile := if (valtype(printFile) == "C", printFile, "" )
   CrLf      := if (valtype(DOSline  ) == "L" .and. DOSline, ;
                    chr(13) + chr(10), chr(10) )
#Cinline
   {
#include <malloc.h>

#if !defined(A_CHARTEXT) && defined(FGSAIX)
#define A_CHARTEXT 0xffff
#endif

      extern variable *_bb_savescreen();                /* if not used yet */
      extern variable *_bb_maxcol();                    /* if not used yet */

      variable *actscreen = VAR_NEW;
      WINDOW   *winPtr;
      int      x, y, xmax;
      unsigned char *oneline, *str;

      actscreen = _bb_savescreen(0,0);                    /* SAVE SCREEN   */
      winPtr    = VAR_SCR (actscreen);
      xmax      = VAR_INTNUM (fsFnNAME(maxcol)(0,0) );    /* xmax=MAXCOL() */

      oneline = (FSuchar *) malloc((unsigned)((xmax + 1)
											  * sizeof(unsigned char)));
      if (oneline != NULL) {
         for (y=0; y <= ymax; y++) {
            str = oneline;
            for (x=0; x <= xmax; x++)
               *(str++) = (unsigned char)
                          (mvwinch(winPtr,y,x) & A_CHARTEXT);    /* curses */
            SET_VAR_CHRLEN (VAR_ARRELEM (VAR_NAME_LOCAL(chr_arr), y+1),
                            oneline, xmax +1);
         }
         free (oneline);
      }
   }
#endCinline

   SET (_SET_CONSOLE, .F.)                       // set console OFF
   if !empty(printFile)
      SET (_SET_PRINTFILE, printfile)            // set printer TO ..
   endif
   SET (_SET_PRINTER, .T.)                       // set printer ON

   ii := 0
   qqout (CrLf)
   if printRow
**        ---- PrintScreen mm/dd/yy hh:mm:ss ----+---------+---------+
**    00: Content of line 0 ...
**    01: Content of line 1 ...
**      : .....
**    24: Content of line 24...
**        ---------+---------+---------+---------+---------+---------+
      dummy := "---- PrintScreen " + dtoc(date()) + " " + time() + " ----+"
      qqout (SPACE(4) + dummy + REPLICATE("---------+", 4) + CrLf)
      AEVAL (chr_arr, {|x| QQOUT(STRZERO(ii++,2) + ":", x + CrLf)} )
      qqout (SPACE(4) + REPLICATE("---------+", 8) + CrLf)
   else
      AEVAL (chr_arr, {|x| QQOUT(x) + CrLf} )
   endif

   SET (_SET_CONSOLE, lSetCons)                  // set console <orig on/off>
   SET (_SET_PRINTER, lSetPrint)                 // set printer <orig on/off>
   if !empty(cSetPrint)
      SET(_SET_PRINTFILE, cSetPrint)             // set printer <orig file>
   endif
   SETPOS (nRow, nCol)                           // reset cursor

   RETURN NIL

* Ŀ
*  printstat        printer status, simulated only for Unix                
* 
function printstat(nDev)
   local ret := 0
   if !file("/dev/lp" + str(nDev -1,1))
      ret := 8
   endif
   return ret

********************************************
* Functions to Access drives               *
********************************************


* Ŀ
*  diskspeed                        Unix not a real time operating system, 
*                                   so there is no fixed speed index       
* 
function diskspeed
parameters p1
fs2_devWarn(MSG_HARDW + "1000")
return 1000         && This is 10 times the speed of an XT

* Ŀ
*  diskfree                                              Same as Diskspace 
* 
function diskfree(p1)
return diskspace(p1)   // std. FS function

* Ŀ
*  diskready                if there is a path, that drive is always ready 
* 
function diskready
parameters p1, p2
fs2_devWarn(MSG_HARDW + ".T.")
return .T.          && return "ready"

* Ŀ
*  diskreadyw               if there is a path, that drive is always ready 
* 
function diskreadyw
parameters p1, p2
fs2_devWarn(MSG_HARDW + ".T.")
return .T.          && return "ready"


********************************************
* Functions to access novell networks      *
********************************************


* Ŀ
*  nnetsdate                                 Our date IS the server's date 
* 
function nnetsdate
return date()

* Ŀ
*  nnetstime                                  Our time IS the servers time 
* 
function nnetstime
return time()

* Ŀ
*  nnetwhoami                               The login name of this process 
* 
function nnetwhoami
return gete("LOGNAME")

* Ŀ
*  nnetinuse                     Actually, we have processed, not Stations 
* 
function nnetinuse
return 32

* Ŀ
*  nnetstanum                                    All users are on one host 
* 
function nnetstanum
fs2_devWarn(MSG_NOVEL + "netname() value")
return netname()

* Ŀ
*  nnetstaid                                     All users are on one host 
* 
function nnetstaid
fs2_devWarn(MSG_NOVEL + "netname() value")
return netname()

* Ŀ
*  nnetocnums                                    All users are on one host 
* 
function nnetocnums
parameters p1c
fs2_devWarn(MSG_NOVEL + "netname() value")
return netname()

* Ŀ
*  nnetwork              Is a novell net active?                           
*  Since used for testing the network (which is always active, ret = .T.   
* 
function nnetwork
fs2_devWarn(MSG_NOVEL + ".T.")
return .T.

***********************************************************
* The com_*() functions to access serial (or other) ports *
* are now in the fs2_com.prg file                         *
***********************************************************

***********************************************************
* Functions to keep compatibility                         *
***********************************************************

* Ŀ
*  osver                 Same as std.funct. OS()                           
* 
function osver
return os()

* Ŀ
*  getsecret             simulated, not 1:1 to CA3                         
* 
function getsecret(cVal, nRow, nCol, lSay, cSay)           // simulated
   local getlist := {}, curColor := setcolor()
   cVal := if (valtype(cVal) != "C", space(10), cVal)
   nRow := if (valtype(nRow) != "N", row(), nRow)
   nCol := if (valtype(nCol) != "N", col(), nCol)
   lSay := if (valtype(lSay) != "L", .F.,   lSay)
   cSay := if (valtype(cSay) != "C", NIL,   cSay)

   if cSay != NIL
      @ nRow, nCol SAY cSay
      nCol := col() +1
   endif
   setcolor ("N/N,N/N")
   @ nRow, nCol GET cVal
   READ
   setcolor (curColor)
   if lSay
      @ nRow, nCol SAY cVal
   endif
   return cVal

******************************************************************************
* fs2_devWarn(cText)
*
* reports Developer Warning when FS_SET("devel", .T.) is set
*-----------------------------------------------------------------------------
#include "error.fh"

FUNCTION fs2_devWarn(cText)
local oDefErr, oErr
if fs_set("devel")
   oDefErr := ErrorBlock()                  // default error handler block
   oErr    := ErrorNew()                    // new error object

   oErr:ARGS          := NIL
   oErr:CANDEFAULT    := .F.
   oErr:CANRETRY      := .F.
   oErr:CANSUBSTITUTE := .F.
   oErr:CARGO         := NIL
   oErr:DESCRIPTION   := PROCNAME(1) + "(): " + cText
   oErr:FILENAME      := PROCNAME(2) + "(" + LTRIM(STR(PROCLINE(2),5)) + ")"
   oErr:GENCODE       := EG_DEVWARNING
   oErr:OPERATION     := ""
   oErr:OSCODE        := 0
   oErr:SEVERITY      := ES_RTERR
   oErr:SUBCODE       := 0
   oErr:SUBSYSTEM     := "BASE"

   EVAL (oDefErr, oErr)
endif
return NIL

* eof
