/* A program to test the Postgres95 - FlagShip API.
 *
 * Browse the data in the part table of the abc database.
 * Edit the quantity of the chosen record.
 *
 * To COMPILE: FlagShip pgtest.prg SQLkit_PG.o -o pgtest -lpq
 */

#include "fileio.fh"
#include "inkey.fh"

// Retrieved records (i.e., tmp files) will be written in this directory.
// Make certain this directory already exits and that the user that runs
// this program has read/write permission in this directory. Be sure the
// name ends with a slash as shown below.
//                                      Edit to suit your system 
//                                      here.
FS_SET("setenvir", "PG_RETRIEVE_DIR", "/home/postgres/.retrieve/")

// Errors will be appended to this file.
FS_SET("setenvir", "PG_ERROR_LOG", "/home/postgres/.pg_errors")

// Open a database and check for errors.
LOCAL nRetVal := open_pg( "abc" )
IF .NOT. nRetVal == 0
  ? "Error opening the abc database"
  WAIT "Hit any key to exit."
  QUIT
ENDIF

// This is the select statement used below.
LOCAL cSelect := "select * from part order by part_no"

// Initialize variables used below.
LOCAL cTfilename := ""
LOCAL nFieldCount := 0
LOCAL nRowCount := 0

// Retrieve the records and check for errors.
nRowCount := ret_pg(cSelect, @cTfilename, @nFieldCount)
IF .NOT. nRowCount > 0
  ? "Error retrieving data."
  WAIT "Hit any key to exit."
  QUIT
ENDIF

// Variables used to read the data from the temporary file, etc.
LOCAL text, handle
LOCAL_INT nC1, nC2, nBarPos

// This two-dimensional array will hold each data element retrieved.
LOCAL aBrow [nRowCount] [nFieldCount]

// read the temporary file
handle := FOPEN( cTfilename, FO_READ )
FOR nC1 := 1 TO nRowCount
  // Read a line of text equivalent to a single record.
  text := FREADTXT( handle )
  // Parse each record into its separate fields; each field ends with a bar.
  FOR nC2 := 1 TO nFieldCount
      nBarPos := AT( "|", text )
      aBrow[nC1][nC2] := SUBSTR( text, 1, nBarPos-1 )
      text := SUBSTR( text, nBarPos+1 )
  NEXT
NEXT

FCLOSE( handle )             /* close the temporary file */
FERASE( cTfilename )         /* delete the temporary file */ 

// Get the data into a one-dimensional array of string... easy to 
// display using ACHOICE(). Or, you could create a Browse object and
// load aBrow into it, in which case you would not need
// aMyarray, etc.
LOCAL aMyarray[nRowCount]

FOR nC1 := 1 TO nRowCount
  aMyarray[nC1] := PADR(aBrow[nC1][1], 7) + PADR(aBrow[nC1][2], 23) + ;
        PADR(aBrow[nC1][3], 5)
NEXT

// Display screen starts here.
@ 1, 1 SAY "ABC Auto Parts Co.                                     Browse/Edit parts table"
LOCAL cPCount := STR(LEN(aBrow)) + " Parts" 
@ 3, 3 SAY cPCount

@ 6, 7 SAY "ID     Name                   Quantity"

LOCAL nChoice := ACHOICE(8, 7, 20, 45, aMyarray)

// Since aMyarray and aBrow are parallel, nChoice now holds
// the index of the record you have chosen to update.

// Allow user to update the quantity for the selected record.
BEGIN SEQUENCE
  IF .NOT. nChoice == 0

    // Since we are about to update the chosen record, get a lock on it
    // by starting a transaction block.

    nRetVal = begin_pg()
    IF nRetVal == -1
      ? "Error starting a transaction block."
      WAIT "Hit any key to exit."
      QUIT
    ENDIF

    @ 3, 3 CLEAR TO MAXROW(), MAXCOL() 
    @ 3, 3 SAY aMyarray[nChoice]
    LOCAL cNewQuantity := SPACE(4)
    @ 5, 19 SAY "New Quantity:" GET cNewQuantity PICTURE "@! 9999"
    READ

    // User may ESC without affecting the database.
    IF LASTKEY() == K_ESC
      rback_pg()
      BREAK
    ENDIF

    cNewQuantity := ALLTRIM( cNewQuantity )

    // If the user hits Enter on an empty cNewQuantity field, a poorly
    // formed SQL command would be the result, so an empty field will
    // be interpreted as a zero.
    IF LEN( cNewQuantity ) == 0
      cNewQuantity := "0"
    ENDIF

    // Update the record, check for errors.
    LOCAL cUpdateStr := "update part set quantity = " + cNewQuantity + " where part_no = " + aBrow[nChoice][1]
    nRetVal := exec_pg( cUpdateStr )
    IF nRetVal == -1
      ? "Error during update."
      WAIT "Hit any key to exit."
      rback_pg()
      QUIT
    ENDIF

    // Make the user agree to the update
    DO WHILE .T.
      @ 7, 7 SAY "OK to update the database? (y/n): "
      LOCAL nAnswer := INKEY(0)
      IF CHR(nAnswer) == "y" .OR. CHR(nAnswer) == "Y"
        ?? "Y"
        nRetVal := end_pg()
        IF nRetVal == -1
          ? "Error committing update."
          WAIT "Hit any key to exit."
          QUIT
        ENDIF
        @ 9, 7 SAY "Database updated."
        INKEY(3)
        BREAK
      ENDIF
      IF CHR(nAnswer) == "n" .OR. CHR(nAnswer) == "N" 
        ?? "N"     
        nRetVal := rback_pg()
        IF nRetVal == -1
          ? "Error during rollback of update."
          WAIT "Hit any key to exit."
          QUIT
        ENDIF   
        @ 9, 7 SAY "Database not updated."
        INKEY(3)
        BREAK
      ENDIF
      ?? CHR(7)  /* none of the above, beep at the user */
    ENDDO
  ENDIF
END

// Close the database, and we're done.
nRetVal := close_pg()

// That's all, folks!

