
#include "webkit.fh"

/*

EXAMPLE ROUTE LOOKUP

$Id: example5.prg,v 1.1 1997/09/26 01:47:52 catfood Exp $

WebKit: CGI and FastCGI interface library for FlagShip
Copyright (C) 1996,1997   Mark W. Schumann <webkit@underflap.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Log: example5.prg,v $
Revision 1.1  1997/09/26 01:47:52  catfood
Initial revision


*/

function example5 (aEnv, aForm, oState) as usual
local x1 as numeric // Starting x point
local x2 as numeric // Starting y point
local y1 as numeric // Ending x point
local y2 as numeric // Ending y point
local time1 as string   // Starting time
local stop1 as string  // Starting stop
local stop2 as string  // Ending stop
local lines1 as array // Array of routes to start
local lines2 as array // Array of routes to end

wkhtml()
? "<TITLE>Transit trip planner[tm]</TITLE>"
wkbody()

? "<h1>Transit Trip Planner[tm]</h1>"
? "Current state: " + ltrim (str (oState:state)) + "<BR>"

do case
case oState:state == 0 // startup
	state0 (aEnv, aForm, oState)
case oState:state == 1 // working
case oState:state == 2 // end
	close all
endcase
wkbodyend()
wkhtmlend()
return oState

static function state1 (aEnv, aForm, oState)
x1 := val (wklookup (aForm, "X1"))
y1 := val (wklookup (aForm, "Y1"))
x2 := val (wklookup (aForm, "X2"))
y2 := val (wklookup (aForm, "Y2"))
time1 := wklookup (aForm, "TIME")

stop1 := closestop (x1, y1)
stop2 := closestop (x2, y2)
lines1 := linesbystop (stop1)
lines2 := linesbystop (stop2)

? "The closest stop to your starting point is " + namestop (stop1) + ".<BR>"
? "The closest stop to your destination is " + namestop (stop2) + ".<BR>"
? "You starting point is on the following lines:<BR>"
aeval (lines1, {|z| wkline (nameline (z) + "<BR>")})
? "Your destination is on the following lines:<BR>"
aeval (lines2, {|z| wkline (nameline (z) + "<BR>")})

? "You can get there these ways:<BR>"
showroutes (lines1, lines2, time1)

oState:state := 2
return oState

//
//	Return the descriptive name of the given stop.
//
function namestop (cName)
return iif (stop->(dbseek (cName)), alltrim (stop->stopname), "NO STOP")

//
//	Return the closest stop.
//
function closestop (xcoord, ycoord)
local_int nDist2
local_int nShortest
local cStop

nShortest := 999999
cStop := "ERR"
stop->(dbgotop())
while !stop->(eof())
   nDist2 := (xcoord - stop->xcoord) ** 2 + (ycoord - stop->ycoord) ** 2
   if (nDist2 < nShortest)
      nShortest := nDist2
      cStop := stop->stopid
   endif
   stop->(dbskip())
end
return cStop

//
//	Return a string description of the given route.
//
function nameline (cName)
local cRet
cRet := "NO ROUTE"
if route->(dbseek (cName))
   cRet := "(" + alltrim (cName) + ") " + alltrim (route->routename)
endif
return cRet

//
//	Return an unordered array of all routes passing by a given
//	stop number.
//
function linesbystop (cStop)
local a

a := {}
rtstop->(dbgotop())
while !rtstop->(eof())
   if rtstop->stopid == cStop
      aadd (a, rtstop->routeid)
   endif
   rtstop->(dbskip())
end

return a

//
//	Return TRUE if and only if there is a transfer between
//	the two routes.
//
function istransfer (cRoute1, cRoute2)
return trx->(dbseek (cRoute1 + cRoute2))

//
//	Return an array of all routes that appear in both arrays.
//
function showroutes (aRoutes1, aRoutes2, cTime1)

for i := 1 to len (aRoutes1)
   for j := 1 to len (aRoutes2)
      if (aRoutes1[i] == aRoutes2[j])
         wkline (nameline (aRoutes1[i]) + " <EM>DIRECT</EM>")
      endif
      if istransfer (aRoutes1[i], aRoutes2[j])
         wkline (nameline (aRoutes1[i]) + ;
            " transfer to " + nameline (aRoutes2[j]))
      endif
   next
next

return nil

function reindex

use stop exclusive
index on stopid to stop1
index on str(xcoord,3,0)+str(ycoord,3,0) to stop2
index on str(ycoord,3,0)+str(xcoord,3,0) to stop2

use route exclusive
index on routeid to route1

use rtstop exclusive
index on routeid+stopid to rtstop1

use trx exclusive
index on route1+route2 to trx1
index on route2+route1 to trx2

use stoptime
index on routeid+tripid+seq to stoptim1

close databases

return nil


static function state0 (aEnv, aForm, oState)
reindex()
use stop     new index stop1, stop2, stop3
use route    new index route1
use rtstop   new index rtstop1
use trx      new index trx1
use stoptime new index stoptim1
? "<HR>"
? "<FORM ACTION=" + thisurl (aEnv) + " METHOD=POST>"
? "Starting point: <INPUT TYPE=STRING><BR>"
? "Destination point: <INPUT TYPE=STRING><BR>"
? "<INPUT TYPE=SUBMIT NAME=Submit><BR>"
? "<INPUT TYPE=RESET NAME=Forget><BR>"
? "</FORM>"
? "<HR>"
? "<PRE>"
? "Transit Trip Planner[tm]"
? "Last updated " + dtoc (date())
oState:state := 1
return nil

function thisurl (aEnv)
return "http://" + wklookup (aEnv, "HTTP_HOST") + wklookup (aEnv, "SCRIPT_NAME") + wklookup