/*

$Id: suferror.prg,v 1.1 1997/09/26 01:47:52 catfood Exp $

WebKit: CGI and FastCGI interface library for FlagShip
Copyright (C) 1996,1997   Mark W. Schumann <webkit@underflap.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Log: suferror.prg,v $
Revision 1.1  1997/09/26 01:47:52  catfood
Initial revision


*/

#define SUBSYSTEM "SUFERROR"

#include "sufmisc.ch"
#include "fileio.ch"

#define LOGFILE "ERROR.LOG"

ANNOUNCE suferror

//============================================================================
//
//   Module SUFERROR
//
//   Revision History
//   Who Date      Comments
//   MWS 18-Feb-96 Added e:filename display
//
//
//============================================================================

//============================================================================
//
//   Function ERRORSYS (NIL) --> NIL
//
//   Purpose:
//      Initialize the error system.  Replaces ERRORSYS() in CLIPPER.LIB.
//
//============================================================================

INIT FUNCTION errorsys
RETURN ERRORBLOCK ({|e| errhand (e)})

//============================================================================
//
//   Function THROWERROR (cSubSystem, nSubCode, nSeverity, aArgs) --> NIL
//
//   Purpose:
//      Create an Error class object and toss it to the current
//      error handler.
//
//
//   THIS MODULE IS AN UNPUBLISHED TRADE SECRET OF MARK W. SCHUMANN.
//   UNAUTHORIZED MODIFICATION, DISTRIBUTION, OR DISCLOSURE OF THIS
//   SOURCE CODE OR THE RESULTING OBJECT OR BINARY CODE IS PROHIBITED.
//
//
//============================================================================

FUNCTION throwerror (cSubSystem, nSubCode, nSeverity, cDesc, axArgs)
LOCAL b := errorblock()
LOCAL e := errornew()
DEFAULT nSeverity TO ES_ERROR
DEFAULT cDesc TO "<no description available>"
e:args := aclone (axArgs)
e:subsystem := cSubSystem
e:subcode := nSubCode
e:severity := nSeverity
e:description := cDesc
eval (b, e)
RETURN NIL


//============================================================================
//
//   Function ERRHAND (err) --> lHandled
//
//   Purpose:
//      Display useful information for run-time errors.
//
//   Mark W. Schumann
//
//============================================================================

FUNCTION errhand (e)
LOCAL bOldErr
LOCAL aInfo
LOCAL lRet
LOCAL i

bOldErr := ERRORSYS()
aInfo := {}
lRet := .F.

IF e:severity > ES_WHOCARES
    AADD (aInfo, "An error has occurred.")
 
    AADD (aInfo, "Subsystem:   " + e:subsystem)
    AADD (aInfo, "Subcode:     " + ALLTRIM (STR (e:subcode)))
    AADD (aInfo, "Description: " + e:description)
    IF !EMPTY (DOSERROR())
        AADD (aInfo, "DOS Error:   " + ALLTRIM (STR (DOSERROR())))
    ENDIF
    IF !EMPTY (e:operation)
        AADD (aInfo, "Operation:   " + e:operation)
    ENDIF
    IF !EMPTY (e:filename)
        AADD (aInfo, "File:        " + e:filename)
    ENDIF
    IF VALTYPE (e:args) == "A" .AND. LEN (e:args) > 0
        AADD (aInfo, "Arguments:")
        FOR i := 1 TO LEN (e:args)
            AADD (aInfo, ;
             " [" + ALLTRIM (STR (i)) + "] " + ;
              VALTYPE (e:args[i]) + " {" + X2C (e:args[i]) + "}")
        NEXT
    ENDIF
    i := 0
    AADD (aInfo, "Call stack:")
    WHILE !EMPTY (PROCNAME (i+2))
        AADD (aInfo, " [" + ALLTRIM (STR (i)) + "] " + ;
         PROCNAME (i+2) + " (" + ALLTRIM (STR (PROCLINE (i+2))) + ")")
        i++
    END
    IF (showerr (e:subSystem, aInfo) > 0)
        dumperr (aInfo)
    ENDIF
    CLEAR SCREEN
    QUIT
    lRet := .F.
ELSE
    AADD (aInfo, e:description)
    //showerr (aInfo)
    lRet := .T.
ENDIF

ERRORBLOCK (bOldErr)
RETURN lRet


#ifdef CLIP4WIN

//============================================================================
//
//   Function SHOWERR (cSubsystem, aInfo)-- Clip4Win version
//
//   Purpose:
//      Display the error dump in a nice window.
//
//   Mark W. Schumann
//
//============================================================================

STATIC FUNCTION showerr (cSubsystem, aInfo)
LOCAL cString
LOCAL nRet

nRet := .F.
cString := ""
AEVAL (aInfo, {|z| cString += z, cString += CRLF})

cString += (CRLF + "Do you want to save error logging info?")
nRet := MessageBox (GetFocus(), cString, cSubSystem, MB_YESNO)

RETURN nRet

#else

//============================================================================
//
//   Function SHOWERR (cSubsystem, aInfo)-- DOS version
//
//   Purpose:
//      Display the error dump in a nice window.
//
//   Mark W. Schumann
//
//============================================================================

STATIC FUNCTION showerr (cSubSystem, aInfo)
LOCAL cScrn
LOCAL cColor
LOCAL nLeft
LOCAL nRight
LOCAL nTop
LOCAL nBottom
LOCAL nWidth
LOCAL i
LOCAL c

nTop := INT ((MAXROW() - LEN (aInfo)) / 2) - 1
RANGE nTop TO 0, MAXROW()

nWidth := 39  // Minimum width--Size of "Do you want...?" prompt.
FOR i := 1 TO LEN (aInfo)
    IF LEN (aInfo[i]) > nWidth
        nWidth := LEN (aInfo[i])
    ENDIF
NEXT
nLeft := INT ((MAXCOL() - nWidth) / 2) - 2
RANGE nLeft TO 0, MAXCOL()

nBottom := nTop + LEN (aInfo) + 3
RANGE nBottom TO 0, MAXROW()

nRight := nLeft + nWidth + 3
RANGE nRight TO 0, MAXCOL()

cScrn := SAVESCREEN (nTop, nLeft, nBottom, nRight)
cColor := SETCOLOR (IIF (ISCOLOR(), "W+/R", "W/N"))

@ nTop, nLeft CLEAR TO nBottom, nRight
@ nTop, nLeft TO nBottom, nRight DOUBLE
FOR i := 1 TO LEN (aInfo)
    @ nTop + i, nLeft + 2 SAY aInfo[i]
NEXT

#ifdef _XYZZY
@ nTop + i + 1, nLeft + 2 SAY "Do you want to save error logging info?"
WHILE (c := UPPER (CHR (INKEY (0))), !(c == 'Y' .or. c == 'N'))
END
#else
@ nTop + i + 1, nLeft + 2 SAY "Please press a key to exit the program."
INKEY (0)
#endif

SETCOLOR (cColor)

RESTSCREEN (nTop, nLeft, nBottom, nRight, cScrn)

#ifdef _XYZZY
RETURN IIF (c == 'Y', 1, 0)
#else
RETURN 1
#endif


#endif

//============================================================================
//
//   Function DUMPERR (aInfo) --> NIL
//
//   Purpose:
//      Dump error-logging information to a file.
//
//   Mark W. Schumann
//
//============================================================================

STATIC FUNCTION dumperr (aInfo)
LOCAL f

f := FOPEN (LOGFILE, FO_WRITE)
IF f <= 0
    f := FCREATE (LOGFILE)
ELSE
    FSEEK (f, 0, FS_END)
ENDIF

FWRITE (f, "Date: " + DTOC (DATE()) + " " + TIME() + CRLF)
AEVAL (aInfo, {|z| FWRITE (f, z + CRLF) })
FWRITE (f, REPLICATE ("-", 77) + CRLF)
FCLOSE (f)

RETURN NIL

//============================================================================
//
//   Function X2C (xVal) --> cVal
//
//   Purpose:
//      Return character equivalent of any data type.
//
//   Mark W. Schumann
//
//============================================================================

STATIC FUNCTION x2c (xVal)
LOCAL cVal
LOCAL cType := VALTYPE (xVal)
DO CASE
    CASE cType == "N"
        cVal := ALLTRIM (STR (xVal))
    CASE cType == "C"
        cVal := xVal
    CASE cType == "L"
        cVal := IIF (xVal, ".T.", ".F.")
    CASE cType == "D"
        cVal := DTOC (xVal)
    CASE cType == "O"
        cVal := "<object>"
    CASE cType == "A"
        cVal := "<array>"
    CASE cType == "B"
        cVal := "<block>"
    CASE cType == NIL
        cVal := "NIL"
    OTHERWISE
        cVal := "<unknown>"
ENDCASE
RETURN cVal

