/*

ERROR HANDLER INVOKED BY THE FLAGSHIP RUN-TIME ENGINE

$Id: wkerr.prg,v 1.2 1997/09/30 04:45:38 catfood Exp catfood $

WebKit: CGI and FastCGI interface library for FlagShip
Copyright (C) 1996,1997   Mark W. Schumann <webkit@underflap.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Log: wkerr.prg,v $
Revision 1.2  1997/09/30 04:45:38  catfood
Multiple fixes to bring us to WK release 1.02.

Revision 1.1  1997/09/26 01:47:52  catfood
Initial revision


*/

#Cinline
#include "_webkit.h"
#endCinline

#define SUBSYSTEM "SUFERROR"

#include "error.fh"
#include "webkit.fh"

/*

	errhand (e) --> <lResponse>

	The error handler invoked by the FlagShip
	run-time engine.

*/

//============================================================================
//
//   Function ERRORSYS (NIL) --> NIL
//
//   Purpose:
//      Initialize the error system.  Replaces ERRORSYS() in CLIPPER.LIB.
//
//============================================================================

FUNCTION errorsys
wksyslog (LOG_DEBUG, "error handler set")
RETURN ERRORBLOCK ({|e| errhand (e)})

//============================================================================
//
//   Function THROWERROR (cSubSystem, nSubCode, nSeverity, aArgs) --> NIL
//
//   Purpose:
//      Create an Error class object and toss it to the current
//      error handler.
//
//
//   THIS MODULE IS AN UNPUBLISHED TRADE SECRET OF MARK W. SCHUMANN.
//   UNAUTHORIZED MODIFICATION, DISTRIBUTION, OR DISCLOSURE OF THIS
//   SOURCE CODE OR THE RESULTING OBJECT OR BINARY CODE IS PROHIBITED.
//
//
//============================================================================

FUNCTION throwerror (cSubSystem, nSubCode, nSeverity, cDesc, axArgs)
LOCAL b := errorblock()
LOCAL e := errornew()
#ifdef FlagShip
DEFAULT nSeverity TO ES_RTERR
DEFAULT cDesc TO "<no description available>"
#else
DEFAULT nSeverity TO ES_ERROR
DEFAULT cDesc TO "<no description available>"
#endif
e:args := aclone (axArgs)
e:subsystem := cSubSystem
e:subcode := nSubCode
e:severity := nSeverity
e:description := cDesc
eval (b, e)
RETURN NIL


//============================================================================
//
//   Function ERRHAND (err) --> lHandled
//
//   Purpose:
//      Display useful information for run-time errors.
//
//   Mark W. Schumann
//
//============================================================================

FUNCTION errhand (e)
LOCAL bOldErr
LOCAL aInfo
LOCAL lRet
LOCAL i

//#ifndef FlagShip
bOldErr := ERRORSYS()
//#endif
aInfo := {}
lRet := .F.

#ifndef FlagShip
IF e:severity > ES_WHOCARES
#endif
    AADD (aInfo, "An error has occurred.")
 
    AADD (aInfo, "Subsystem:   " + e:subsystem)
    AADD (aInfo, "Subcode:     " + ALLTRIM (STR (e:subcode)))
    AADD (aInfo, "Description: " + e:description)
    IF !EMPTY (DOSERROR())
        AADD (aInfo, "DOS Error:   " + ALLTRIM (STR (DOSERROR())))
    ENDIF
    IF !EMPTY (e:operation)
        AADD (aInfo, "Operation:   " + e:operation)
    ENDIF
    IF !EMPTY (e:filename)
        AADD (aInfo, "File:        " + e:filename)
    ENDIF
    IF VALTYPE (e:args) == "A" .AND. LEN (e:args) > 0
        AADD (aInfo, "Arguments:")
        FOR i := 1 TO LEN (e:args)
            AADD (aInfo, ;
             " [" + ALLTRIM (STR (i)) + "] " + ;
              VALTYPE (e:args[i]) + " {" + X2C (e:args[i]) + "}")
        NEXT
    ENDIF
    i := 0
    AADD (aInfo, "Call stack:")
    WHILE !EMPTY (PROCNAME (i+2))
        AADD (aInfo, " [" + ALLTRIM (STR (i)) + "] " + ;
         PROCNAME (i+2) + " (" + ALLTRIM (STR (PROCLINE (i+2))) + ")")
        i++
    END
    dumperr (aInfo)
#ifndef WEBKIT_VERSION
    CLEAR SCREEN
#endif
    QUIT
    lRet := .F.
#ifndef FlagShip
ELSE
    AADD (aInfo, e:description)
    //showerr (aInfo)
    lRet := .T.
ENDIF
#endif

ERRORBLOCK (bOldErr)
RETURN lRet


//============================================================================
//
//   Function DUMPERR (aInfo) --> NIL
//
//   Purpose:
//      Dump error-logging information to a file.
//
//   Mark W. Schumann
//
//============================================================================

STATIC FUNCTION dumperr (aInfo)
#ifdef WEBKIT_VERSION
AEVAL (aInfo, {|z| wksyslog (LOG_ERR, z) })
AEVAL (aInfo, {|z| qout (z + "<BR>" + EOL)})
#else
AEVAL (aInfo, {|z| qout (z + EOL) })
#endif
RETURN NIL

//============================================================================
//
//   Function X2C (xVal) --> cVal
//
//   Purpose:
//      Return character equivalent of any data type.
//
//   Mark W. Schumann
//
//============================================================================

FUNCTION x2c (xVal)
LOCAL cVal
LOCAL cType := VALTYPE (xVal)
DO CASE
    CASE cType == "N"
        cVal := ALLTRIM (STR (xVal))
    CASE cType == "C"
        cVal := xVal
    CASE cType == "L"
        cVal := IIF (xVal, ".T.", ".F.")
    CASE cType == "D"
        cVal := DTOC (xVal)
    CASE cType == "O"
        cVal := "<object>"
    CASE cType == "A"
        cVal := "<array>"
    CASE cType == "B"
        cVal := "<block>"
    CASE cType == NIL
        cVal := "NIL"
    OTHERWISE
        cVal := "<unknown>"
ENDCASE
RETURN cVal

