
/*

$Id: wkmain.c,v 1.2 1997/09/30 04:45:38 catfood Exp catfood $

WebKit: CGI and FastCGI interface library for FlagShip
Copyright (C) 1995,1996,1997   Mark W. Schumann <webkit@underflap.com>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Log: wkmain.c,v $
Revision 1.2  1997/09/30 04:45:38  catfood
Multiple fixes to bring us to WK release 1.02.

Revision 1.1  1997/09/26 01:47:52  catfood
Initial revision


*/

#include <FSopenc.h>    /* access to FS variables from C */
#include <FSextend.h>   /* FS functions written in C */
#include "_webkit.h"    /* internal WK functions */
#include "util.h"       /* NCSA CGI header file */

#include <sys/types.h> /* for testing */
#include <sys/stat.h> /* for testing */
#include <fcntl.h> /* for testing */
#include <string.h>
#include <unistd.h>

char *fmakeword (FILE *, char, int *);
extern FSvar *_bb_wkstatenew FGS_ANSI((FSint, FSvar *[]));
char *makeword (char *, char);

static variable *state = 0;
static int count = 0;

extern char **environ;
 
UDF_PROT(webkitmain);
UDF_PROT(wkdirect);
UDF_PROT(aadd);
UDF_PROT(cursesinit);
UDF_PROT(programini);
UDF_PROT(usersmax);
UDF_PROT(fs_set);
/*

	main

	If the program is being called via CGI, we interpret
	the form variables and environment variables, then
	pass them to the webkitmain() program supplied by the
	user.

*/


#ifdef USE_FASTCGI

#include "fcgi_stdio.h" /* for FastCGI */

FSudfname (main)

{
int fa;

	FSinit();

	openlog ("webkit", LOG_CONS, LOG_LOCAL0);
	syslog (LOG_DEBUG, "using FastCGI");
	syslog (LOG_DEBUG, "compiled on " __DATE__ " at " __TIME__);

	while ((fa = FCGI_Accept()) >= 0)
	{
	FSvar *v[1];
	int n = 0;

		syslog (LOG_DEBUG, "FCGI_Accept() returns %d", fa);
		FCGI_SetExitStatus (n);
		VAR_NEW_ARGS (v, 1);
		_bb_wkdirect (1, v);
	}
	syslog (LOG_DEBUG, "FCGI_Accept() returns %d", fa);
	FCGI_Finish();
	syslog (LOG_DEBUG, "It's done. Stick a fork in it.");
	_ret();
	FSreturn;
}

#else

FSudfname (main)

{
FSvar *v[1];

	FSinit();

	openlog ("webkit", LOG_CONS, LOG_LOCAL0);
	syslog (LOG_DEBUG, "using regular CGI");
	syslog (LOG_DEBUG, "compiled on " __DATE__ " at " __TIME__);
	VAR_NEW_ARGS (v, 1);
	_bb_wkdirect (1, v);
	syslog (LOG_DEBUG, "It's done. Stick a fork in it.");
	_ret();
	FSreturn;
}

#endif

FSudfname (wkdirect)
{

variable *aEnv;
variable *aForm;
int envcount;
char **s;               /* Pointer into environment area */
int i;
char *eqsign;

	FSinit();
	syslog (LOG_DEBUG, "PID %u; iteration #%d", getpid(), count++);

	for (s = environ, envcount = 0; *s; s++)
	{
		++envcount;
	}
	envcount += 3;

	VAR_NEW_ARRAY (aEnv, envcount);
	syslog (LOG_DEBUG, "created aEnv[%d] at %p", envcount, aEnv);

	/* Initialize the array to elements of {NIL, NIL} */
	for (i = 1; i <= envcount; i++)
	{
	variable *aTemp;

		SET_VAR_COPY (
		 VAR_ARRELEM (aEnv, i), VAR_NEW_ARRAY (aTemp, 2));
	}

	for (i = 0, s = environ; i < envcount && *s; i++, s++)
	{
		eqsign = (char *) strchr (*s, '=');
		if (eqsign == NULL)
		{
		}
		else
		{
			SET_VAR_CHRLEN (
			 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 1),
			 *s, eqsign - *s);
			SET_VAR_CHR (
			 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 2),
			 eqsign + 1);
		}
	}

	/* Setting up WebKit's predefined environment variables. */
	{
	uid_t uid;
	struct passwd *pw;

		uid = getuid();
		pw = getpwuid (uid);
		syslog (LOG_DEBUG, "user is %s", pw->pw_name);
		SET_VAR_CHR (
		 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 1), "WEBKIT_USER");
		SET_VAR_CHR (
		 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 2), pw->pw_name);
		++i;
	}

	SET_VAR_CHR (
	 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 1), "WEBKIT_RELEASE");
	SET_VAR_CHR (
	 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 2), "1.10");
	++i;

	SET_VAR_CHR (
	 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 1), "WEBKIT_PLATFORM");
	SET_VAR_CHR (
	 VAR_ARRELEM (VAR_ARRELEM (aEnv, i + 1), 2), "Linux");
	++i;

	syslog (LOG_DEBUG, "done with environment variables");

	/* Now create an array of form variables. */
	{
	register int x,m=0;
	int cl;
	char *clstring;

		clstring = getenv ("CONTENT_LENGTH");
		syslog (LOG_DEBUG, "content-length is (%p) \"%s\"",
		 clstring, clstring);
		if (clstring == NULL)
		{
			cl = 0;
		}
		else
		{
			cl = atoi (clstring);
		}
		syslog (LOG_DEBUG, "content-length is %d", cl);

		VAR_NEW_ARRAY (aForm, 0);
		syslog (LOG_DEBUG, "created aForm[%d] at %p", 0, aForm);

		for (x = 0; cl && (!feof (stdin)); x++)
		{
		variable *aTemp;
		variable *args[2];
		variable *ret;
		char *entval;
		char *entname;

			m = x;
			entval = fmakeword (stdin, '&', &cl);
			plustospace (entval);
			unescape_url (entval);
			entname = makeword (entval, '=');
			syslog (LOG_DEBUG,
			 "form variable \"%s\" is \"%s\"", entname, entval);
			VAR_NEW_ARRAY (aTemp, 2);
			SET_VAR_CHR (VAR_ARRELEM (aTemp, 1), entname);
			SET_VAR_CHR (VAR_ARRELEM (aTemp, 2), entval);
			VAR_NEW_ARGS (args, 2);
			SET_VAR_COPY (args[0], aForm);
			SET_VAR_COPY (args[1], aTemp);
			ret = _bb_aadd (2, args);
			free (entval);
			free (entname);
		}

	}

	writenl ("Content-type: text/html");
	writenl ("");

	{
	variable *args[3];
	variable *v;

		VAR_NEW_ARGS (args, 3);
		args[0] = aEnv;
		args[1] = aForm;
		syslog (LOG_DEBUG, "there are %d envars", VAR_ISDIM (aEnv));
		syslog (LOG_DEBUG, "there are %d formvars", VAR_ISDIM (aForm));
		if (state == 0)
		{
			state = _bb_wkstatenew (0, (variable **) 0);
		}
		args[2] = state;
		syslog (LOG_DEBUG, "calling webkitmain()");
		v = _bb_webkitmain (3, args);
	}

	_FSret();
	FSreturn;
}

extern void fgsUse4html();

UDF_DECL(cursesinit)
{
    FSvar *myargc[3];	

#ifdef __linux__
    if (VAR_INT(UDF_EXEC(usersmax)(0,0)) != 4) 
	   setenv ("TERM", "dumb", 1);
#else
    if (VAR_INT(UDF_EXEC(usersmax)(0,0)) != 4) {
	   VAR_NEW_ARGS(myargc, 3);
	   SET_VAR_CHR(myargc[0], "setenvir");
	   SET_VAR_CHR(myargc[1], "TERM");
	   SET_VAR_CHR(myargc[2], "dumb");
	   UDF_EXEC(fs_set)(3, myargc);
	}
#endif
	return NIL_VAR;
}

#ifndef USEOWNINIT
FSudfname (programini)
{
	fgsUse4html();
}
#endif

FSudfname (fastcgi)
{
	FSinit();
#ifdef FASTCGI
	_FSretl (1);
#else
	_FSretl (0);
#endif
	FSreturn;
}

